###############################################################################
# Licensed Materials - Property of IBM
# (c) Copyright IBM Corporation 2011, 2016. All Rights Reserved.
#
# Note to U.S. Government Users Restricted Rights:  
# Use, duplication or disclosure restricted by GSA ADP Schedule 
# Contract with IBM Corp. 
###############################################################################

#
# Script that installs the Green Hat jdbc provider into WebSphere Application Server
#

import os
import re
import sys
import shutil
import javax.management as management
import com.ibm.ws.management.filetransfer.client as options

#############################################################
#                                                           #
# Parameter checking and validation of the install location #
#                                                           #
#############################################################

def usage():
  sys.exit("\nUsage: wsadmin.bat -f setupJDBCProvider.py <install directory>\n" +
             "       | installNode <node name>\n"
             "       | installNode <node name> installServer <server name>\n"
             "       | installCluster <cluster name>\n\n" +
             "The <install directory> should include the jdbc driver files, and is required.\n\n" +
             "With no other arguments the driver will be installed to every server, " +
             "node, and cluster in the WebSphere Application Server cell. To limit " + 
             "the installation to a specific node, server, or cluster, provide either " +
             "a node name, node and server names, or a cluster name.\n\n" +
             "Please use forward slashes ('/') in the install directory path.")
#endDef

selectedNode = 0
selectedServer = 0
selectedCluster = 0

argc = len(sys.argv)
if argc == 0 or argc == 2 or argc == 4 or argc > 5:
  usage()

if argc == 3:
  if sys.argv[1] == "installNode":
    selectedNode = sys.argv[2]
  elif sys.argv[1] == "installCluster":
    selectedCluster = sys.argv[2]
  else:
    usage()

if argc == 5:
  if sys.argv[1] == "installNode" and sys.argv[3] == "installServer":
    selectedNode = sys.argv[2]
    selectedServer = sys.argv[4]
  else:
    usage()

# Any input such as '\t' will have been mapped into a tab character by the time that
# we get the path, so we need to reverse that mapping. This is not perfect, as user
# strings including '\x' or '\\' will still be mapped incorrectly or may fail. If
# there are problems then it will help to use forward slashes ('/') in the path.
raw_input = sys.argv[0]
gh_install_path = ""
for ch in raw_input:
  if ch == '\a':
    gh_install_path = gh_install_path + '/a'
  elif ch == '\b':
    gh_install_path = gh_install_path + '/b'
  elif ch == '\f':
    gh_install_path = gh_install_path + '/f'
  elif ch == '\n':
    gh_install_path = gh_install_path + '/n'
  elif ch == '\r':
    gh_install_path = gh_install_path + '/r'
  elif ch == '\t':
    gh_install_path = gh_install_path + '/t'
  elif ch == '\v':
    gh_install_path = gh_install_path + '/v'
  elif ch == '\\':
    gh_install_path = gh_install_path + '/'
  else:
    gh_install_path = gh_install_path + ch

if not gh_install_path.endswith('/'):
  gh_install_path = gh_install_path + '/'

required_files = [
    'com.ibm.rational.rit.jdbcdriver4.jar',
    'greenhat.jdbcdriver.jar',
    'jsqlparser15.jar',
    'registration.xml'
]

for file in required_files:
  file = gh_install_path + file
  try:
    fh = open(file, 'r')
    fh.close
  except IOError:
    sys.exit('\nERROR: Failed to find required file at path: ' + file)

####################
#                  #
# Helper functions #
#                  #
####################

global AdminConfig
global AdminControl
global driverVariableName

driverVariableName = "GH_DRIVER_PATH"

def setupDriverPathVariable(scope, install_path):
  variableMap = AdminConfig.getid(scope + "VariableMap:/")

  found = 0
  variables = AdminConfig.list("VariableSubstitutionEntry",variableMap).splitlines()
  for variable in variables:
    name = AdminConfig.showAttribute(variable, "symbolicName")
    if name == driverVariableName:
      print "\tUpdating existing WebSphere variable " + driverVariableName
      AdminConfig.modify(variable,[["value", install_path]])
      found = 1
      break
    
  if not found:
    print "\tCreating new WebSphere variable " + driverVariableName
    AdminConfig.create("VariableSubstitutionEntry",variableMap,[["symbolicName",driverVariableName],["value",install_path]])
#endDef

def findProfileLocation(node):
  variableMap = AdminConfig.getid("/Node:"+node+"/VariableMap:/")
  variables = AdminConfig.list("VariableSubstitutionEntry",variableMap).splitlines()
  for variable in variables:
    name = AdminConfig.showAttribute(variable, "symbolicName")
    if name == 'USER_INSTALL_ROOT':
      return AdminConfig.showAttribute(variable, "value")
#endDef

def upload(transfer, localFile, dmgrFile, options):
  file = java.io.File(localFile)
  AdminControl.invoke_jmx(transfer, 'uploadFile',\
  [file, dmgrFile, options],\
  ['java.io.File','java.lang.String','com.ibm.websphere.management.filetransfer.client.FileTransferOptions'])
#endDef

def download(transfer, dmgrFile, destFile, options):
  file = java.io.File(destFile)
  AdminControl.invoke_jmx(transfer, 'downloadFile',\
  [dmgrFile, file, options],\
  ['java.lang.String','java.io.File','com.ibm.websphere.management.filetransfer.client.FileTransferOptions'])
#endDef

#
# Setup the Green Hat JDBC provider. The parent is provided to try and find the set
# of servers that can use the provider, so that it is created with the correct settings
# for that version of WAS.
# 
def setupProvider(parent, scopeName, scope):
  print "  " + scopeName

  version61 = 0
  
  servers = AdminConfig.list('Server', parent).splitlines()
  clusterMembers = AdminConfig.list('ClusterMember', parent).splitlines()
  if len(servers) == 0:
    if len(clusterMembers) > 0:
      servers = []
      for member in clusterMembers:
        node = AdminConfig.showAttribute(member, 'nodeName')
        server = AdminConfig.showAttribute(member, 'memberName')
        servers.append(AdminConfig.getid('/Node:'+node+'/Server:'+server+'/'))
    else:
      servers = [parent]

  for server in servers:
    match = re.search('/nodes/(?P<node>[^/]+)/servers/(?P<server>[^|]+)', server)
    if match:
      nodeName = match.group('node')
      serverName = match.group('server')
      server = AdminControl.completeObjectName("WebSphere:*,type=Server,node=" + nodeName + ",name=" + serverName)
      if server.find(',version=6.1') != -1:
        version61 = 1
        break

  if version61:
    print "\tUsing WAS 6.1 style configuration"

  providerName = "Green Hat JDBC Provider"
  driverClassPath = ""
  if version61:
    driverClassPath += "${" + driverVariableName + "}greenhat.jdbcdriver.jar;"
  else:
    driverClassPath += "${" + driverVariableName + "}com.ibm.rational.rit.jdbcdriver4.jar;"

  driverClassPath += "${" + driverVariableName + "}jsqlparser15.jar;"

  properties = [
    ["name",providerName],
    ["implementationClassName","com.greenhat.jdbc.XADataSource"],
    ["classpath",driverClassPath],
    ["xa","true"],
    ["description","Green Hat JDBC Provider. This provider enables database stubbing with IBM Rational Integration Tester."]
  ]

  if not version61:
    properties.append(["isolatedClassLoader","true"])

  provider = AdminConfig.getid(scope + "JDBCProvider:"+ providerName + "/")

  # Switch over to the id for the current scope
  scope = AdminConfig.getid(scope)

  if version61 and provider:
    canDelete = 1
    dataSources = AdminConfig.list("DataSource",scope).splitlines()
    for dataSource in dataSources:
      dataSourceProvider = AdminConfig.showAttribute(dataSource, 'provider')
      if dataSourceProvider == provider:
        canDelete = 0
        sys.exit('\nERROR: Unable to remove existing Green Hat JDBC Provider as it is in use. \
Please disable JDBC virtualization for all data sources at this scope and retry.')
        break
    if canDelete:
      print "\tDeleting existing JDBC provider"
      AdminConfig.remove(provider)
      provider = 0

  if not provider:
    print "\tCreating new JDBC provider"
    properties.append(["providerType","User-defined JDBC Provider"])
    provider = AdminConfig.create("JDBCProvider",scope,properties)
  elif not version61:
    print "\tUpdating existing JDBC provider"
    AdminConfig.resetAttributes(provider,properties)
#endDef

###########################
#                         #
# Main part of the script #
#                         #
###########################

#
# Detect if there is a dmgr controlling the cell
#
try:
  connectedNode = AdminControl.getNode()
except:
  print("\nERROR: Please start the WebSphere processes before running this script.\n")
  raise

connectedServer = AdminControl.queryNames("node="+connectedNode+",type=Server,*")
dmgr = AdminControl.getAttribute(connectedServer, 'processType') == "DeploymentManager"

#
# Decide which cell, nodes, clusters and servers to install onto
#
cell = java.lang.System.getenv('WAS_CELL')
nodes = AdminConfig.list("Node").splitlines()
clusters = AdminConfig.list("ServerCluster").splitlines()
servers = AdminConfig.list("Server").splitlines()

if selectedCluster and not dmgr:
  sys.exit("\nERROR: WebSphere clusters are only supported when working with a WAS Deployment Manager")
elif selectedCluster:
  nodes = []
  servers = []
  for candidate in clusters:
    name = AdminConfig.showAttribute(candidate, 'name')
    if name == selectedCluster:
      clusters = [candidate]
      members = AdminConfig.list('ClusterMember', candidate).splitlines()
      for member in members:
        nodeName = AdminConfig.showAttribute(member, 'nodeName')
        node = AdminConfig.getid('/Node:'+nodeName+'/')
        serverName = AdminConfig.showAttribute(member, 'memberName')
        server = AdminConfig.getid('/Node:'+nodeName+'/Server:'+serverName+'/')
        servers.append(server)
        try:
          nodes.index(node)
        except:
          # The node is not yet in the list
          nodes.append(node)
      break
  if len(nodes) == 0:
    sys.exit("\nERROR: Cannot find cluster " + selectedCluster)

elif selectedServer and not dmgr:
  sys.exit("\nERROR: Server selection is only supported when working with a WAS Deployment Manager")
elif selectedServer:
  clusters = []
  nodes = [AdminConfig.getid('/Node:'+selectedNode+'/')]
  servers = [AdminConfig.getid('/Node:'+selectedNode+'/Server:'+selectedServer+'/')]
  if len(nodes[0]) == 0:
    sys.exit("\nERROR: Cannot find node " + selectedNode)
  if len(servers[0]) == 0:
    sys.exit("\nERROR: Cannot find server " + selectedServer + " in node " + selectedNode)

elif selectedNode and not dmgr:
  sys.exit("\nERROR: Node selection is only supported when working with a WAS Deployment Manager")
elif selectedNode:
  clusters = []
  nodes = [AdminConfig.getid('/Node:'+selectedNode+'/')]
  if len(nodes[0]) == 0:
    sys.exit("\nERROR: Cannot find node " + selectedNode)
  servers = AdminConfig.list('Server', nodes[0]).splitlines()

#
# Check that we have work to do
#
if len(nodes) == 0:
  sys.exit("\nERROR: There are no WebSphere nodes to configure")

#
# Whenever a dmgr is in play we use file transfer services to push the driver to all of the nodes
#
localNode = java.lang.System.getenv('WAS_NODE')
if dmgr:
  targetPath = 0
  options = options.FileTransferOptionsImpl()
  options.setOverwrite(java.lang.Boolean.TRUE)
  options.setDeleteSourceOnCompletion(java.lang.Boolean.FALSE)
  
  if localNode != connectedNode:
    # If we are not local then we need to upload the files to the dmgr
    print "\nUploading the Green Hat JDBC driver from " + gh_install_path + " to the Deployment Manager node"
    try:
      localTransfer = AdminControl.completeObjectName('WebSphere:*,node=' + localNode + ',type=FileTransferClient')
      localTransfer = management.ObjectName(localTransfer)
    except:
      sys.exit("\nERROR: Please start the file transfer service on node: " + localNode)

    try:
      for file in required_files:
        upload(localTransfer, gh_install_path + file, file, options)
    except:
      print("\nERROR: Failed to upload the files to the Deployment Manager. " \
          + "Re-running the installation using wsadmin within on the Deployment Manager node is recommended.")
      raise
      
  else:
    # We already have access to the dmgr filesystem, so we can copy the files directly
    print "\nCopying the Green Hat JDBC driver from " + gh_install_path + " to the Deployment Manager node"
    targetPath = findProfileLocation(localNode) + "/config/temp/"
    for file in required_files:
      shutil.copyfile(gh_install_path + file, targetPath + file)
  
  # Download the driver to the targets
  print "\nDownloading the Green Hat JDBC driver to the Application Server nodes:"
  for targetNode in nodes:
    if targetNode == nodes[len(nodes)-1] and not targetPath:
      options.setDeleteSourceOnCompletion(java.lang.Boolean.TRUE)
      
    name = AdminConfig.showAttribute(targetNode, "name")
    if name != connectedNode:
      print "  " + name
      if name == localNode:
        remoteTransfer = localTransfer
      else:
        try:
          remoteTransfer = AdminControl.completeObjectName('WebSphere:*,node=' + name + ',type=FileTransferClient')
          remoteTransfer = management.ObjectName(remoteTransfer)
        except:
          sys.exit("\nERROR: Please start the file transfer service on node: " + name)

      # Copy the files down into the profile
      profilePath = findProfileLocation(name) + '/RIT-Platform/'
      for file in required_files:
        download(remoteTransfer, file, profilePath + file, options)
      # Set a WebSphere variable to point to the file location
      setupDriverPathVariable("/Node:" + name + "/", '${USER_INSTALL_ROOT}/RIT-Platform/')
    elif connectedNode == localNode:
      # We are connected directly to the dmgr
      print "\n  " + localNode
      setupDriverPathVariable("/Cell:" + cell + "/", gh_install_path)
    else:
      # We are remotely connected to the dmgr, so we cannot move the files there
      print("\nWARNING: Unable to copy the driver files onto the Deployment Manager node. " \
          + "JDBC data sources at the cell level may be unable to test connections to the database. " \
          + "Re-running the installation using wsadmin within on the Deployment Manager node is recommended");

  if targetPath:
    # Delete the files that we copied into the temp directory
    try:
      for file in required_files:
        os.unlink(targetPath + file)
    except:
      print("\nWARNING: Failed to clean up temporary files. Please remove platform pack files from " + targetPath)

else:
  # On a standalone server we don't copy the files around, we just point to the existing location
  # We set the variable at both cell and node scope as the old installer used node scope and we need
  # to overwrite the out of date location.
  if len(nodes) > 1:
    sys.exit("\nERROR: More than one node cannot be configured without a Deployment Manager")
  print "\nSetting " + driverVariableName + " onto cell " + cell
  setupDriverPathVariable("/Cell:" + cell + "/", gh_install_path)
  
  name = AdminConfig.showAttribute(nodes[0], "name")
  print "\nSetting " + driverVariableName + " onto node " + name
  setupDriverPathVariable("/Node:" + name + "/", gh_install_path)

#
# Setup JDBC providers on each target scope
#

primaryNode = AdminConfig.getid("/Node:" + connectedNode + "/")
if dmgr:
  dmgrServer = AdminConfig.list('Server',primaryNode)
  try:
    servers.remove(dmgrServer)
  except:
    # nothing to do
    servers = servers

print "\nInstalling Green Hat JDBC provider onto:"
setupProvider(primaryNode, "Cell " + cell, "/Cell:" + cell + "/")

for cluster in clusters:
  clusterName = AdminConfig.showAttribute(cluster, "name")
  setupProvider(cluster, "  Cluster " + clusterName, "/ServerCluster:" + clusterName + "/")

for node in nodes:
  nodeName = AdminConfig.showAttribute(node, "name")
  setupProvider(node, "  Node " + nodeName, "/Node:" + nodeName + "/")
  
  nodeAppServers = AdminConfig.list("ApplicationServer", node).splitlines()
  for server in servers:
    appServer = AdminConfig.list("ApplicationServer", server)
    try:
      # This will throw an exception if the server is not in the list
      nodeAppServers.index(appServer)
      serverName = AdminConfig.showAttribute(server, "name")
      setupProvider(server, "    Server " + serverName, "/Node:" + nodeName + "/Server:" + serverName + "/")
    except:
      # Do nothing
      server = server

AdminConfig.save()
print "\nSaving configuration changes"

if dmgr:
  print "Synchronizing active nodes"
  sync = AdminControl.completeObjectName('WebSphere:*,type=DeploymentManager')
  AdminControl.invoke(sync, 'multiSync', 'true')

print "\nGreen Hat JDBC Provider installation completed successfully"
